{
  "nbformat": 4,
  "nbformat_minor": 0,
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.8.6"
    },
    "colab": {
      "name": "BooleanRetrieval.ipynb",
      "provenance": [],
      "collapsed_sections": []
    }
  },
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "duq4gOzIpp7F"
      },
      "source": [
        "# A Simple Boolean Retrieval System"
      ]
    },
    {
      "cell_type": "code",
      "metadata": {
        "id": "MnJXnmpdpp7J"
      },
      "source": [
        "from functools import total_ordering, reduce\n",
        "import csv\n",
        "import re"
      ],
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "ty7C6ESQpp7L"
      },
      "source": [
        "### Postings"
      ]
    },
    {
      "cell_type": "code",
      "metadata": {
        "id": "qBhikJmwpp7L"
      },
      "source": [
        "@total_ordering\n",
        "class Posting:\n",
        "    \n",
        "    def __init__(self, docID):\n",
        "        self._docID = docID\n",
        "        \n",
        "    def get_from_corpus(self, corpus):\n",
        "        return corpus[self._docID]\n",
        "    \n",
        "    def __eq__(self, other):\n",
        "        return self._docID == other._docID\n",
        "    \n",
        "    def __gt__(self, other):\n",
        "        return self._docID > other._docID\n",
        "    \n",
        "    def __repr__(self):\n",
        "        return str(self._docID)"
      ],
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "lLw9bNHLpp7M"
      },
      "source": [
        "### Posting Lists"
      ]
    },
    {
      "cell_type": "code",
      "metadata": {
        "id": "u81tZ4y3pp7M"
      },
      "source": [
        "class PostingList:\n",
        "    \n",
        "    def __init__(self):\n",
        "        self._postings = []\n",
        "    \n",
        "    @classmethod\n",
        "    def from_docID(cls, docID):\n",
        "        plist = cls()\n",
        "        plist._postings = [(Posting(docID))]\n",
        "        return plist\n",
        "    \n",
        "    @classmethod\n",
        "    def from_posting_list(cls, postingList):\n",
        "        plist = cls()\n",
        "        plist._postings = postingList\n",
        "        return plist\n",
        "    \n",
        "    def merge(self, other):\n",
        "        i = 0\n",
        "        last = self._postings[-1]\n",
        "        while (i < len(other._postings) and last == other._postings[i]):\n",
        "            i += 1\n",
        "        self._postings += other._postings[i:]\n",
        "        \n",
        "    def intersection(self, other):\n",
        "        intersection = []\n",
        "        i = 0\n",
        "        j = 0\n",
        "        while (i < len(self._postings) and j < len(other._postings)):\n",
        "            if (self._postings[i] == other._postings[j]):\n",
        "                intersection.append(self._postings[i])\n",
        "                i += 1\n",
        "                j += 1\n",
        "            elif (self._postings[i] < other._postings[j]):\n",
        "                i += 1\n",
        "            else:\n",
        "                j += 1\n",
        "        return PostingList.from_posting_list(intersection)\n",
        "    \n",
        "    def union(self, other):\n",
        "        union = []\n",
        "        i = 0\n",
        "        j = 0\n",
        "        while (i < len(self._postings) and j < len(other.postings)):\n",
        "            if (self._postings[i] == other._postings[j]):\n",
        "                union.append(self._postings[i])\n",
        "                i += 1\n",
        "                j += 1\n",
        "            elif (self._postings[i] < other._postings[j]):\n",
        "                union.append(self._postings[i])\n",
        "                i += 1\n",
        "            else:\n",
        "                union.append(other._postings[j])\n",
        "                j += 1\n",
        "        for k in range(i, len(self._postings)):\n",
        "            union.append(self._postings[k])\n",
        "        for k in range(j, len(other._postings)):\n",
        "            union.append(other._postings[k])\n",
        "        return PostingList.from_posting_list(union)\n",
        "    \n",
        "    def get_from_corpus(self, corpus):\n",
        "        return list(map(lambda x: x.get_from_corpus(corpus), self._postings))\n",
        "    \n",
        "    def __repr__(self):\n",
        "        return \", \".join(map(str, self._postings))"
      ],
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "5wU2rlGXpp7N"
      },
      "source": [
        "### Terms"
      ]
    },
    {
      "cell_type": "code",
      "metadata": {
        "id": "fxYhdnn6pp7O"
      },
      "source": [
        "class ImpossibleMergeError(Exception):\n",
        "    pass\n",
        "\n",
        "@total_ordering\n",
        "class Term:\n",
        "    \n",
        "    def __init__(self, term, docID):\n",
        "        self.term = term\n",
        "        self.posting_list = PostingList.from_docID(docID)\n",
        "        \n",
        "    def merge(self, other):\n",
        "        if (self.term == other.term):\n",
        "            self.posting_list.merge(other.posting_list)\n",
        "        else:\n",
        "            raise ImpossibleMergeError\n",
        "    \n",
        "    def __eq__(self, other):\n",
        "        return self.term == other.term\n",
        "    \n",
        "    def __gt__(self, other):\n",
        "        return self.term > other.term\n",
        "    \n",
        "    def __repr__(self):\n",
        "        return self.term + \": \" + repr(self.posting_list)"
      ],
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "T-_DNEW8pp7P"
      },
      "source": [
        "### Inverted Index"
      ]
    },
    {
      "cell_type": "code",
      "metadata": {
        "id": "Y2gETwNMpp7Q"
      },
      "source": [
        "def normalize(text):\n",
        "    no_punctuation = re.sub(r'[^\\w^\\s^-]','',text)\n",
        "    downcase = no_punctuation.lower()\n",
        "    return downcase\n",
        "\n",
        "def tokenize(movie):\n",
        "    text = normalize(movie.description)\n",
        "    return list(text.split())\n",
        "\n",
        "class InvertedIndex:\n",
        "    \n",
        "    def __init__(self):\n",
        "        self._dictionary = []\n",
        "        \n",
        "    @classmethod\n",
        "    def from_corpus(cls, corpus):\n",
        "        intermediate_dict = {}\n",
        "        for docID, document in enumerate(corpus):\n",
        "            tokens = tokenize(document)\n",
        "            for token in tokens:\n",
        "                term = Term(token, docID)\n",
        "                try:\n",
        "                    intermediate_dict[token].merge(term)\n",
        "                except KeyError:\n",
        "                    intermediate_dict[token] = term\n",
        "            if (docID % 1000 == 0):\n",
        "                print(\"ID: \" + str(docID))\n",
        "        idx = cls()\n",
        "        idx._dictionary = sorted(intermediate_dict.values())\n",
        "        return idx\n",
        "    \n",
        "    def __getitem__(self, key):\n",
        "        for term in self._dictionary:\n",
        "            if term.term == key:\n",
        "                return term.posting_list\n",
        "        raise KeyError\n",
        "        \n",
        "    def __repr__(self):\n",
        "        return \"A dictionary with \" + str(len(self._dictionary)) + \" terms\""
      ],
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "9V_-ECH8pp7Q"
      },
      "source": [
        "### Reading the Corpus"
      ]
    },
    {
      "cell_type": "code",
      "metadata": {
        "id": "ItpQZcODpp7R"
      },
      "source": [
        "class MovieDescription:\n",
        "    \n",
        "    def __init__(self, title, description):\n",
        "        self.title = title\n",
        "        self.description = description\n",
        "        \n",
        "    def __repr__(self):\n",
        "        return self.title\n",
        "    \n",
        "def read_movie_descriptions():\n",
        "    filename = 'plot_summaries.txt'\n",
        "    movie_names_file = 'movie.metadata.tsv'\n",
        "    with open(movie_names_file, 'r') as csv_file:\n",
        "        movie_names = csv.reader(csv_file, delimiter='\\t')\n",
        "        names_table = {}\n",
        "        for name in movie_names:\n",
        "            names_table[name[0]] = name[2]\n",
        "    with open(filename, 'r') as csv_file:\n",
        "        descriptions = csv.reader(csv_file, delimiter='\\t')\n",
        "        corpus = []\n",
        "        for desc in descriptions:\n",
        "            try:\n",
        "                movie = MovieDescription(names_table[desc[0]], desc[1])\n",
        "                corpus.append(movie)\n",
        "            except KeyError:\n",
        "                pass\n",
        "        return corpus"
      ],
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "XWuqzHRTpp7R"
      },
      "source": [
        "### Putting it all together"
      ]
    },
    {
      "cell_type": "code",
      "metadata": {
        "id": "VwGSkoq9pp7S"
      },
      "source": [
        "class IRsystem:\n",
        "    \n",
        "    def __init__(self, corpus, index):\n",
        "        self._corpus = corpus\n",
        "        self._index = index\n",
        "        \n",
        "    @classmethod\n",
        "    def from_corpus(cls, corpus):\n",
        "        index = InvertedIndex.from_corpus(corpus)\n",
        "        return cls(corpus, index)\n",
        "    \n",
        "    def answer_query(self, words): # ['cat', 'batman']\n",
        "        norm_words = map(normalize, words)\n",
        "        postings = map(lambda w: self._index[w], norm_words)\n",
        "        plist = reduce(lambda x, y: x.intersection(y), postings)\n",
        "        return plist.get_from_corpus(self._corpus)\n",
        "    \n",
        "def query(ir, text):\n",
        "    words = text.split()\n",
        "    answer = ir.answer_query(words)\n",
        "    for movie in answer:\n",
        "        print(movie)"
      ],
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "code",
      "metadata": {
        "id": "51emy_vhpp7T"
      },
      "source": [
        "corpus = read_movie_descriptions()"
      ],
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "code",
      "metadata": {
        "id": "2zNDrhAipp7S"
      },
      "source": [
        "ir = IRsystem.from_corpus(corpus)"
      ],
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "code",
      "metadata": {
        "id": "oQWNPt-rpp7T"
      },
      "source": [
        "query(ir, \"gandalf frodo\")"
      ],
      "execution_count": null,
      "outputs": []
    },
    {
      "cell_type": "code",
      "metadata": {
        "id": "0P0z8Agypp7T"
      },
      "source": [
        ""
      ],
      "execution_count": null,
      "outputs": []
    }
  ]
}